import rich_click as click
from arkitekt_next.cli.vars import get_manifest
import subprocess
from click import Context
from arkitekt_next.cli.io import get_builds
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL


@click.command()
@click.option("--build", help="The build to use", type=str, default=None)
@click.option("--flavour", "-f", help="The flavour to use", default="vanilla")
@click.option(
    "--url",
    "-u",
    help="The fakts server to use",
    type=str,
    default=DEFAULT_ARKITEKT_URL,
)
@click.option(
    "--builder",
    help="The builder to use",
    type=str,
    default="arkitekt_next.builders.easy",
)
@click.pass_context
def stage(ctx: Context, build: str, url: str, flavour: str, builder: str) -> None:
    """Stages the latest Build for testing

    Stages the current build for testing. This will create a temporary staged version
    of the app that is run agains the local arkitekt_next instance. The builder will be changed
    to the easy or provided builder to ensure that the app can be run headlessly


    """

    get_manifest(ctx)

    builds = get_builds(build)

    if len(builds) == 0:
        raise click.ClickException("Could not find any builds")

    if len(builds) > 1:
        try:
            build_model = next(
                build for build in builds.values() if build.flavour == flavour
            )
        except StopIteration:
            raise click.ClickException(
                f"Could not find a build for flavour {flavour}. Please run `arkitekt_next port build` "
                + "first to build the flavour"
            )

    else:
        build_model = list(builds.values())[0]

    command = (
        build_model.build_docker_command()
        + build_model.build_arkitekt_next_command(url)
    )

    click.echo(f"Staging build {build} with flavour {flavour} against {url}")
    click.echo("Running command: " + " ".join(command))

    subprocess.run(" ".join(command), shell=True)

    raise click.ClickException("Docker container exited")



""" Small extension to tqdm that reports progress to arkitekt_next through the
assignation context"""

from tqdm import tqdm as _tqdm
from rekuest_next.actors.vars import current_assignation_helper


class tqdm(_tqdm):
    """A tqdm that reports progress to arkitekt_next through the
    assignation context

    This tqdm assigns the current progress to the current assignation helper
    if it exists. This allows the progress to be reported to the user
    through the ArkitektNext UI.

    TODO: Check if this works with the current and next versions of tqdm. Maybe
    we should factor this out into the rekuest package

    """

    def __init__(self, *args, **kwargs) -> None:
        """The tqdm constructor"""
        super().__init__(*args, **kwargs)

        self._assignationhelper = current_assignation_helper.get(None)

        self.last_arkitekt_next_perc = 0

    def update(self, *args, **kwargs):
        """An update method that reports progress to arkitekt_next through the
        assignation context and the current assignation helper

        Returns
        -------
        The return value of tqdm.update
        """
        z = super().update(*args, **kwargs)
        if self._assignationhelper:
            if self.last_arkitekt_next_perc + 0.05 < self.last_print_n / self.total:
                self.last_arkitekt_next_perc = self.last_print_n / self.total
                self._assignationhelper.progress(
                    int(self.last_arkitekt_next_perc * 100)
                )

        return z



import secrets

from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from arkitekt_next.bloks.base import BaseArkitektService
from typing import List


DEFAULT_ARKITEKT_URL = "http://localhost:8000"


@blok(
    "live.arkitekt.fluss", description="A service for managing workflows and their data"
)
class FlussBlok(BaseArkitektService):
    def __init__(self) -> None:
        self.dev = False
        self.host = "fluss"
        self.command = "bash run-debug.sh"
        self.image = "jhnnsrs/fluss:next"
        self.repo = "https://github.com/jhnnsrs/fluss-server-next"
        self.scopes = {"read_image": "Read image from the database"}
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media"]
        self.secret_key = secrets.token_hex(16)
        self.ensured_repos: List[str] = []

    def get_builder(self):
        return "arkitekt.generic"

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)



from importlib.metadata import version
from arkitekt_next.cli.constants import compile_dockerfiles
from arkitekt_next.cli.types import Flavour
from arkitekt_next.cli.utils import build_relative_dir
import rich_click as click
from click import Context
from rich import get_console
from arkitekt_next.cli.vars import get_manifest
from arkitekt_next.utils import create_arkitekt_next_folder, create_devcontainer_file
import yaml
from rich.panel import Panel

try:
    pass
except ImportError as e:
    raise ImportError("Please install rekuest to use this feature") from e

import os


@click.command()
@click.option("--flavour", "-f", help="The flavour to use", default="vanilla")
@click.option(
    "--description",
    "-d",
    help="The description for this flavour to use",
    default="This is a vanilla flavour",
)
@click.option(
    "--overwrite",
    "-o",
    help="Should we overwrite the existing Dockerfile?",
    is_flag=True,
    default=False,
)
@click.option(
    "--template",
    "-t",
    help="The dockerfile template to use",
    default="vanilla",
    type=click.Choice(compile_dockerfiles()),
)
@click.option(
    "--devcontainer",
    "-d",
    help="Shouwld we create a devcontainer.json file?",
    is_flag=True,
    default=False,
)
@click.option(
    "--arkitekt-version",
    "-av",
    help="Which Arkitekt-version should we use to mount in the container?",
    default=None,
    type=str,
)
@click.pass_context
def init(
    ctx: Context,
    description: str,
    overwrite: bool,
    flavour: str,
    template: str,
    devcontainer: bool,
    arkitekt_version: str = None,
) -> None:
    """Runs the port wizard to generate a dockerfile to be used with port"""

    arkitekt_next_folder = create_arkitekt_next_folder()

    flavour_folder = os.path.join(arkitekt_next_folder, "flavours", flavour)
    if os.path.exists(flavour_folder) and not overwrite:
        raise click.ClickException(
            f"The flavour {flavour} does already exist. Please initialize a different flavour or use the --overwrite flag"
        )
    else:
        os.makedirs(flavour_folder, exist_ok=True)

    config_file = os.path.join(flavour_folder, "config.yaml")
    dockerfile = os.path.join(flavour_folder, "Dockerfile")

    manifest = get_manifest(ctx)

    fl = Flavour(
        selectors=[],
        description=description,
        dockerfile="Dockerfile",
    )


    try:
        package_version = arkitekt_version or version('arkitekt_next')
        print(f"Detected Arkitekt Package version: {package_version}")
    except:
        raise click.ClickException("Could not detect the Arkitekt package version (maybe you are running a dev version). Please provide it with the --arkitekt-version flag")
        


    with open(config_file, "w") as file:
        yaml.dump(fl.dict(), file)

    with open(build_relative_dir("dockerfiles", f"{template}.dockerfile"), "r") as f:
        dockerfile_content = f.read()

    with open(dockerfile, "w") as f:
        f.write(dockerfile_content.format(__arkitekt_version__=package_version))

    if devcontainer or click.confirm("Do you want to create a devcontainer.json file?"):
        create_devcontainer_file(manifest, flavour, dockerfile)

    panel = Panel(
        title=f"Created new flavour [bold]{flavour}[/bold]\n",
        renderable="You can now edit the Dockerfile and add selectors to the config.yaml file\n"
        + "To learn more about selectors and how flavours work, please visit [link=https://arkitekt_next.live]https://arkitekt_next.live[/link]",
        style="green",
    )

    get_console().print(panel)



from herre.fakts.grant import FaktsGrant
from herre.fakts.registry import GrantRegistry, GrantType
from herre.grants.oauth2.authorization_code import AuthorizationCodeGrant
from herre.grants.oauth2.redirecters.aiohttp_server import AioHttpServerRedirecter
from herre.grants.oauth2.client_credentials import ClientCredentialsGrant


ARKITEKT_GRANT_REGISTRY = GrantRegistry()


def build_authorization_code_grant(**kwargs) -> FaktsGrant:
    return AuthorizationCodeGrant(**kwargs, redirecter=AioHttpServerRedirecter())


def build_client_credentials_grant(**kwargs) -> FaktsGrant:
    return ClientCredentialsGrant(**kwargs)


ARKITEKT_GRANT_REGISTRY.register_grant(
    GrantType.AUTHORIZATION_CODE,
    build_authorization_code_grant,
)

ARKITEKT_GRANT_REGISTRY.register_grant(
    GrantType.CLIENT_CREDENTIALS,
    build_client_credentials_grant,
)



import rich_click as click
from .dev import dev
from .prod import prod


@click.group()
@click.pass_context
def run(ctx):
    """Runs your arkitekt_next app

    Running your app locally is the first step to developing your app. You can run your app in
    development mode, which will automatically reload your app when you change the code, or in
    production mode, which not reload your app when you change the code, but allows you to
    scale your app to multiple processes.



    """


run.add_command(dev, "dev")
run.add_command(prod, "prod")



from blok import blok, InitContext, Option
from blok.tree import YamlFile, Repo
from typing import Any, Protocol
from blok.utils import check_protocol_compliance
from dataclasses import asdict
from arkitekt_next.bloks.services import (
    GatewayService,
    DBService,
    RedisService,
    S3Service,
    ConfigService,
    MountService,
    AdminService,
    SecretService,
    LokService,
)

from blok.bloks.services.dns import DnsService


class DefaultService(Protocol):
    dev: bool
    repo: str
    command: str
    service_name: str
    host: str
    buckets: list[str]
    scopes: dict[str, str]
    secret_key: str
    mount_repo: bool
    build_repo: bool

    def get_blok_meta(self) -> str: ...

    def get_builder(self) -> str: ...


def create_default_service_dependencies():
    return [
        DnsService,
        GatewayService,
        DBService,
        RedisService,
        S3Service,
        ConfigService,
        MountService,
        AdminService,
        SecretService,
        LokService,
    ]


class BaseArkitektService:
    def get_additional_config(self):
        return {}

    def preflight(
        self,
        init: InitContext,
        lok: LokService,
        dns: DnsService,
        db: DBService,
        gateway: GatewayService,
        redis: RedisService,
        s3: S3Service,
        config: ConfigService,
        mount: MountService,
        admin: AdminService,
        secret: SecretService,
        mount_repo: bool = False,
        host: str = "",
        image: str = "",
        secret_key: str = "",
        build_repo: bool = False,
        command: str = "",
        repo: str = "",
        disable: bool = False,
        dev: bool = False,
    ):
        deps = init.dependencies

        lok.register_scopes(self.scopes)

        path_name = self.host

        gateway_access = gateway.expose(path_name, 80, self.host)

        postgress_access = db.register_db(self.host)
        redis_access = redis.register()
        lok_access = lok.retrieve_credentials()
        admin_access = admin.retrieve()
        minio_access = s3.create_buckets(self.buckets)
        lok_labels = lok.retrieve_labels(
            self.get_blok_meta().service_identifier, self.get_builder()
        )

        dns_result = init.get_service(DnsService).get_dns_result()

        csrf_trusted_origins = []
        for hostname in dns_result.hostnames:
            csrf_trusted_origins.append(f"http://{hostname}")
            csrf_trusted_origins.append(f"https://{hostname}")

        configuration = YamlFile(
            **{
                "db": asdict(postgress_access),
                "django": {
                    "admin": asdict(admin_access),
                    "debug": True,
                    "hosts": ["*"],
                    "secret_key": self.secret_key,
                },
                "redis": asdict(redis_access),
                "lok": asdict(lok_access),
                "s3": asdict(minio_access),
                "scopes": self.scopes,
                "force_script_name": path_name,
                "csrf_trusted_origins": csrf_trusted_origins,
                **self.get_additional_config(),
            }
        )

        config_mount = init.get_service(ConfigService).register_config(
            f"{self.host}.yaml", configuration
        )

        depends_on = []

        if redis_access.dependency:
            depends_on.append(redis_access.dependency)

        if postgress_access.dependency:
            depends_on.append(postgress_access.dependency)

        if minio_access.dependency:
            depends_on.append(minio_access.dependency)

        service = {
            "labels": lok_labels,
            "volumes": [f"{config_mount}:/workspace/config.yaml"],
            "depends_on": depends_on,
        }

        if mount_repo or dev:
            mount = init.get_service(MountService).register_mount(self.host, Repo(repo))
            service["volumes"].extend([f"{mount}:/workspace"])

        if build_repo or dev:
            mount = init.get_service(MountService).register_mount(self.host, Repo(repo))
            service["build"] = mount
        else:
            service["image"] = image

        service["command"] = command

        self.service = service

    def get_additional_options(self):
        return []

    def get_options(self):
        return [
            Option(
                subcommand="dev",
                help="Shoud we run the service in development mode (includes withrepo, mountrepo)?",
                default=self.dev,
            ),
            Option(
                subcommand="host",
                help="The name we should use for the host?",
                default=self.host,
            ),
            Option(
                subcommand="disable",
                help="Shoud we disable the service?",
                default=False,
            ),
            Option(
                subcommand="repo",
                help="Which repo should we use when building the service? Only active if build_repo or mount_repo is active",
                default=self.repo,
            ),
            Option(
                subcommand="command",
                help="Which command should we use when building the service?",
                default=self.command,
            ),
            Option(
                subcommand="mount_repo",
                help="Should we mount the repo into the container?",
                type=bool,
                default=self.mount_repo,
            ),
            Option(
                subcommand="build_repo",
                help="Should we build the container from the repo?",
                type=bool,
                default=self.build_repo,
            ),
            Option(
                subcommand="host",
                help="How should the service be named inside the docker-compose file?",
                default=self.host,
            ),
            Option(
                subcommand="secret_key",
                help="The secret key to use for the django service",
                default=self.secret_key,
            ),
            Option(
                subcommand="image",
                help="The image to use for the service",
                default=self.image,
            ),
            *self.get_additional_options(),
        ]



from pydantic import BaseModel
from typing import Dict, Any
from arkitekt_next.bloks.services.mount import MountService
from blok import blok, InitContext, Option, ExecutionContext
from arkitekt_next.bloks.services.gateway import GatewayService
from blok.bloks.services.dns import DnsService
from blok.tree import YamlFile


@blok("live.arkitekt.tailscale", description="Tailscale is a zero config VPN")
class TailscaleBlok:
    def __init__(self, name: str = "arkitekt") -> None:
        self.name = name
        self.disable = False

    def preflight(
        self,
        gateway: GatewayService,
        mount_service: MountService,
        dns: DnsService,
        net_name: str,
        auth_key: str,
        host_name: str,
        disable: bool = False,
    ):
        self.disable = disable
        if self.disable:
            return
        assert auth_key, "You need to provide an auth_key"

        self.caddy_service = gateway.get_internal_host()
        self.caddy_port = gateway.get_http_port()
        self.mount = mount_service.register_mount("tailscale_state", {})
        self.tailnet_name = net_name
        self.auth_key = auth_key
        self.dns_service = dns.get_dns_result()
        self.host_name = host_name

    def build(self, ex: ExecutionContext):
        if self.disable:
            return

        if not self.tailnet_name:
            # Trying to automatically find the hostname
            for i in self.dns_service.hostnames:
                if i.endswith(".ts.net"):
                    self.tailnet_name = i.split(".")[1]
                    break

        service = {
            "image": "jhnnsrs/tailproxy:latest",
            "volumes": [
                f"{self.mount}:/var/lib/tailscale"  # Persist the tailscale state directory
            ],
            "environment": [
                "TS_STATE_DIR=/var/lib/tailscale",
                f"TS_AUTH_KEY={self.auth_key}",
                f"TS_HOSTNAME={self.host_name}",
                f"TS_TAILNET={self.tailnet_name}",
                f"CADDY_TARGET={self.caddy_service}:{self.caddy_port}",
            ],
        }

        ex.docker_compose.set_nested("services", "tailscale_proxy", service)

    def get_options(self):
        return [
            Option(
                subcommand="net_name",
                help="The name of your tailnet",
                type=str,
            ),
            Option(
                subcommand="auth_key",
                help="The auth_key of your tailnet",
                type=str,
            ),
            Option(
                subcommand="host_name",
                help="The hostname of your tailnet",
                default=self.name,
                type=str,
            ),
            Option(
                subcommand="disable",
                help="Should we disable the creation of the tailscale service?",
                default=self.disable,
                type=str,
            ),
        ]



import rich_click as click
from .inspect import inspect
from .scopes import scopes_group
from .version import version
from .wizard import wizard


@click.group()
@click.pass_context
def manifest(ctx) -> None:
    """Updates the manifest of this app

    The manifest is used to describe the app and its rights (scopes) and requirements, to be run on the platform.
    This manifest is used to authenticate the app with the platform establishing its scopes and requirements.



    """
    pass


manifest.add_command(inspect, "inspect")
manifest.add_command(scopes_group, "scopes")
manifest.add_command(version, "version")
manifest.add_command(wizard, "wizard")



from arkitekt_next.apps.types import App, Fakts, Herre, Manifest
from typing import List, Callable, Dict, Any


class QtApp(App):
    """An app that is built with the easy builder"""

    fakts: Fakts
    herre: Herre
    manifest: Manifest
    services: Dict[str, Any]
    hooks: Dict[str, List[Callable]] = {
        "on_start": [],
        "on_stop": [],
        "on_error": [],
        "on_message": [],
        "on_warning": [],
        "on_info": [],
        "on_debug": [],
        "on_enter": [],
    }

    def register_hook(self, hook_name: str, hook: Callable):
        """Register a hook"""
        self.hooks[hook_name].append(hook)

    def run(self):
        """Run the app"""
        self.services["rekuest"].run()

    async def __aenter__(self):
        await super().__aenter__()
        for service in self.services.values():
            await service.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        for service in self.services.values():
            await service.__aexit__(exc_type, exc_value, traceback)



from typing import Any, Dict, Optional

from arkitekt_next.bloks.services.db import DBService, DBCredentials
from arkitekt_next.bloks.services.mount import MountService
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
import pydantic
import namegenerator
import secrets
from blok import blok, InitContext


@blok(
    DBService,
    description="Postgres with Apache AGE",
)
class PostgresBlok(BaseModel):
    host: str = "db"
    port: int = 5432
    skip: bool = False
    password: str = pydantic.Field(default_factory=lambda: secrets.token_hex(16))
    user: str = pydantic.Field(default_factory=lambda: namegenerator.gen(separator=""))
    image: str = "jhnnsrs/daten:next"
    mount_repo: bool = False
    build_repo: bool = False
    repo: str = "https://github.com/arkitektio/daten-server"
    dev: bool = False

    registered_dbs: dict[str, DBCredentials] = {}
    build_image: Optional[Dict[str, Any]] = None

    def get_dependencies(self):
        return []

    def get_identifier(self):
        return "live.arkitekt.postgres"

    def register_db(self, db_name: str) -> DBCredentials:
        if db_name in self.registered_dbs:
            return self.registered_dbs[db_name]
        else:
            access_credentials = DBCredentials(
                password=self.password,
                username=self.user,
                host=self.host,
                port=self.port,
                db_name=db_name,
                dependency=self.host if not self.skip else None,
            )
            self.registered_dbs[db_name] = access_credentials
            return access_credentials

    def preflight(self, init: InitContext, mount: MountService):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.build_image = {
            "environment": {
                "POSTGRES_USER": self.user,
                "POSTGRES_PASSWORD": self.password,
                "POSTGRES_MULTIPLE_DATABASES": "",
            },
            "labels": ["fakts.service=live.arkitekt.postgres"],
        }

        if self.build_repo or self.dev:
            mount = init.get_service(MountService).register_mount(
                self.host, Repo(self.repo)
            )
            self.build_image["build"] = mount
        else:
            self.build_image["image"] = self.image

    def build(self, context: ExecutionContext):

        self.build_image["environment"]["POSTGRES_MULTIPLE_DATABASES"] = ",".join(
            self.registered_dbs.keys()
        )

        context.docker_compose.set_nested(f"services", self.host, self.build_image)

    def get_options(self):
        with_postgres_password = Option(
            subcommand="password",
            help="The postgres password for connection",
            default=self.password,
        )
        with_user_password = Option(
            subcommand="user",
            help="The postgress user_name",
            default=self.user,
        )
        skip_build = Option(
            subcommand="skip",
            help="Should the service not be created? E.g when pointing outwards?",
            default=self.skip,
        )
        with_image = Option(
            subcommand="image",
            help="The image to use for the service",
            default=self.image,
        )
        with_repo = Option(
            subcommand="repo",
            help="The repo to use for the service",
            default=self.repo,
        )
        build_repo = Option(
            subcommand="build_repo",
            help="Should we build the repo?",
            default=self.build_repo,
        )

        mount_repo = Option(
            subcommand="mount_repo",
            help="Should we mount the repo?",
            default=self.mount_repo,
        )

        dev = Option(
            subcommand="dev",
            help="Should we run the service in development mode (includes withrepo, mountrepo)?",
            default=self.dev,
        )

        return [
            with_postgres_password,
            skip_build,
            with_user_password,
            with_image,
            with_repo,
            build_repo,
            mount_repo,
            dev,
        ]



import os
from arkitekt_next.model import Manifest
import json


def create_arkitekt_next_folder(with_cache: bool = True) -> str:
    """Creates the .arkitekt_next folder in the current directory.

    If the folder already exists, it does nothing.
    It automatically creates a .gitignore file, and a .dockerignore file,
    so that the ArkitektNext credential files are not added to git.

    Parameters
    ----------
    with_cache : bool, optional
        Should we create a cache dir?, by default True

    Returns
    -------
    str
        The path to the .arkitekt_next folder.
    """
    os.makedirs(".arkitekt_next", exist_ok=True)
    if with_cache:
        os.makedirs(".arkitekt_next/cache", exist_ok=True)

    gitignore = os.path.join(".arkitekt_next", ".gitignore")
    dockerignore = os.path.join(".arkitekt_next", ".dockerignore")
    if not os.path.exists(gitignore):
        with open(gitignore, "w") as f:
            f.write(
                "# Hiding ArkitektNext Credential files from git\n*.json\n*.temp\ncache/\nservers/"
            )
    if not os.path.exists(dockerignore):
        with open(dockerignore, "w") as f:
            f.write(
                "# Hiding ArkitektNext Credential files from git\n*.json\n*.temp\ncache/\nservers/"
            )

    return os.path.abspath(".arkitekt_next")


def create_devcontainer_file(
    manifest: Manifest,
    flavour: str,
    docker_file_path,
    devcontainer_path=".devcontainer",
) -> None:
    """Creates a devcontainer.json file in the flavour folder.

    Parameters
    ----------
    flavour_folder : str
        The path to the flavour folder.
    """

    os.makedirs(devcontainer_path, exist_ok=True)

    flavour_container = os.path.join(devcontainer_path, flavour)
    os.makedirs(os.path.join(devcontainer_path, flavour), exist_ok=True)

    devcontainer_file = os.path.join(flavour_container, "devcontainer.json")

    devcontainer_content = {}
    devcontainer_content["name"] = f"{manifest.identifier} {flavour} Devcontainer"
    devcontainer_content["build"] = {}
    devcontainer_content["build"]["dockerfile"] = os.path.join("..", docker_file_path)
    devcontainer_content["build"][
        "context"
    ] = "../.."  # This is the root of the project
    devcontainer_content["runArgs"] = ["--network=host"]

    json.dump(devcontainer_content, open(devcontainer_file, "w"), indent=4)



from typing import Dict, Any
import secrets
from arkitekt_next.bloks.services.mount import MountService
from blok import blok, InitContext, ExecutionContext, Option, Command
from blok.bloks.services.vscode import VsCodeService
from blok.tree import Repo, YamlFile


@blok(
    "live.arkitekt.orkestrator",
    description="Orkestrator is the electron-based UI for arkitekt",
)
class OrkestratorBlok:
    def __init__(self) -> None:
        self.dev = False
        self.disable = True
        self.repo = "https://github.com/arkitektio/orkestrator-next"
        self.build_command = ["yarn"]
        self.up_command = ["yarn", "start"]

    def preflight(
        self,
        init: InitContext,
        mount: MountService,
        vscode: VsCodeService | None = None,
    ):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        if self.disable and not self.dev:
            return
        self.mount = mount.register_mount("orkestrator", Repo(self.repo))

        if vscode is not None:
            vscode.register_task(
                "Run Orkestrator",
                "shell",
                "yarn",
                ["start"],
                {"cwd": f"{self.mount}"},
            )
            vscode.register_task(
                "Build Orkestrator",
                "shell",
                "yarn",
                [],
                {"cwd": f"{self.mount}"},
            )

        self.initialized = True

    def build(self, context: ExecutionContext):
        if self.disable and not self.dev:
            return

        context.install_commands.set_nested(
            "orkestrator", Command(self.build_command, cwd=self.mount)
        )
        context.up_commands.set_nested(
            "orkestrator", Command(self.up_command, cwd=self.mount)
        )
        pass

    def get_options(self):
        with_repos = Option(
            subcommand="repo",
            help="The default repo to use for the orkestrator",
            default=self.repo,
        )
        with_disable = Option(
            subcommand="disable",
            help="Should we disable the orkestrator service?",
            default=self.disable,
        )
        with_dev = Option(
            subcommand="dev",
            help="Should we mount orkestrator as dev?",
            default=self.dev,
        )

        return [
            with_dev,
            with_repos,
            with_disable,
        ]



def missing_install(name: str, error: Exception):
    def real_missing_install(*args, **kwargs):
        raise ImportError(
            f"Missing import: {name}. Please install the missing package. "
        ) from error

    return real_missing_install


try:
    from rekuest_next.register import register
    from rekuest_next.agents.hooks import background
    from rekuest_next.agents.hooks import startup
    from rekuest_next.agents.context import context
    from rekuest_next.state.state import state
    from rekuest_next.actors.reactive.api import progress, aprogress
    from rekuest_next.actors.reactive.api import log, alog
except ImportError as e:
    raise e
    register = missing_install("rekuest_next", e)
    background = missing_install("rekuest_next", e)
    startup = missing_install("rekuest_next", e)
    context = missing_install("rekuest_next", e)
    state = missing_install("rekuest_next", e)
    progress = missing_install("rekuest_next", e)
    aprogress = missing_install("rekuest_next", e)
    log = missing_install("rekuest_next", e)
    alog = missing_install("rekuest_next", e)

from .builders import easy, interactive
from .apps.types import App

__all__ = [
    "App",
    "register",
    "easy",
    "interactive",
    "publicqt",
    "jupy",
    "log",
    "alog",
    "progress",
    "aprogress",
    "scheduler",
    "register_structure",
    "group",
    "useGuardian",
    "useInstanceID",
    "useUser",
    "next",
    "state",
    "context",
    "background",
    "startup",
    "register_next",
]



from typing import Dict, Any
import secrets


from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import Repo, YamlFile
from arkitekt_next.bloks.base import BaseArkitektService


@blok("live.arkitekt.kabinet", description="a container and app management service")
class KabinetBlok(BaseArkitektService):
    def __init__(self) -> None:
        self.dev = False
        self.host = "kabinet"
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/arkitektio/kabinet-server"
        self.scopes = {
            "kabinet_deploy": "Deploy containers",
            "kabinet_add_repo": "Add repositories to the database",
        }
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media"]
        self.secret_key = secrets.token_hex(16)
        self.ensured_repos = ["jhnnsrs/ome:main", "jhnnsrs/renderer:main"]
        self.image = "jhnnsrs/kabinet:next"

    def get_additional_config(self):
        return {"ensured_repos": self.ensured_repos}

    def get_builder(self):
        return "arkitekt.generic"

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)

    def get_additional_options(self):
        with_repos = Option(
            subcommand="repos",
            help="The default repos to enable for the service",
            default=self.secret_key,
        )

        return [
            with_repos,
        ]



import rich_click as click
from arkitekt_next.cli.options import *
import asyncio
from arkitekt_next.cli.ui import construct_run_panel
from importlib import import_module
import rich_click as click
from arkitekt_next.cli.options import *
import asyncio
from arkitekt_next.cli.ui import construct_run_panel
from importlib import import_module
from arkitekt_next.cli.utils import import_builder
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL


async def call_app(
    app: App,
    hash,
    arg,
):
    async with app:
        raise NotImplementedError("This is not implemented yet")


@click.command("prod")
@click.option(
    "--url",
    help="The fakts url for connection",
    default=DEFAULT_ARKITEKT_URL,
    envvar="FAKTS_URL",
)
@with_builder
@with_token
@with_instance_id
@with_headless
@with_log_level
@with_skip_cache
@click.pass_context
@click.option(
    "--arg",
    "-a",
    "args",
    help="Key Value pairs for the setup",
    type=(str, str),
    multiple=True,
)
@click.option(
    "--hash",
    "-h",
    help="The hash of the node to run",
    type=str,
)
def remote(ctx, entrypoint=None, builder=None, args=None, hash=str, **builder_kwargs):
    """ALlows you to run a get the output of a node in a remote app.

    This is useful for debugging and testing. In this mode the app itself will not
    be run, so local nodes cannot be called. Only nodes that are availabble on your
    arkitekt_next server can be called.

    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)
    entrypoint = entrypoint or manifest.entrypoint

    kwargs = dict(args or [])

    builder = import_builder(builder)

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    app = builder(
        **manifest.to_builder_dict(),
        **builder_kwargs,
    )

    panel = construct_run_panel(app)
    console.print(panel)

    asyncio.run(call_app(app, hash, kwargs))



import logging
from typing import Optional
import os
import yaml
from arkitekt_next.cli.types import Flavour
from arkitekt_next.cli.vars import get_console
import rich_click as click


def search_username_in_docker_info(docker_info: str):
    for line in docker_info.splitlines():
        if "Username" in line:
            return line.split(":")[1].strip()


def validate_flavours(flavours_folder: str, only: Optional[str]):
    for dir_name in os.listdir(flavours_folder):
        dir = os.path.join(flavours_folder, dir_name)
        if os.path.isdir(dir):
            if only is not None and only != dir_name:
                continue

            if os.path.exists(os.path.join(dir, "config.yaml")):
                with open(os.path.join(dir, "config.yaml")) as f:
                    valued = yaml.load(f, Loader=yaml.SafeLoader)
                try:
                    flavour = Flavour(**valued)

                except Exception as e:
                    get_console().print_exception()
                    raise click.ClickException(f"Flavour {dir_name} is invalid") from e

                try:
                    flavour.check_relative_paths(dir)
                except Exception as e:
                    raise click.ClickException(
                        f"Relative Paths in Flavour {dir_name} are invalid. {e}"
                    ) from e

            else:
                raise click.ClickException(
                    f"Flavour {dir_name} is invalid. No config.yaml file found"
                )

        else:
            logging.info(
                f"Found file {dir_name} in flavours folder. Not a valid flavour. Ignoring"
            )



from arkitekt_next.bloks.services.name import NameService
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from pydantic import BaseModel
from typing import Dict, Any

from blok import blok, InitContext
from blok.bloks.services.dns import DnsService

DEFAULT_PUBLIC_URLS = ["127.0.0.1"]
DEFAULT_PUBLIC_HOSTS = ["localhost"]


class ExposedHost(BaseModel):
    host: str
    port: int
    stip_prefix: bool = True


class ExpostedToHost(BaseModel):
    port: int
    host: str
    to: str
    tls: bool = False


class ExposedPort(BaseModel):
    port: int
    host: str
    tls: bool = False
    to: int


@blok(
    "live.arkitekt.gateway", description="A gateway for exposing services on the host"
)
class GatewayBlok:
    def __init__(self) -> None:
        self.exposed_hosts = {}
        self.exposed_to_hosts = {}
        self.http_expose_default = None
        self.exposed_ports = {}
        self.with_certer = True
        self.with_tailscale = True
        self.certer_image = "jhnnsrs/certer:next"
        self.http_port = 80
        self.https_port = 443
        self.public_ips = DEFAULT_PUBLIC_URLS
        self.public_hosts = DEFAULT_PUBLIC_HOSTS

    def get_internal_host(self):
        return "caddy"

    def get_https_port(self):
        return 443

    def get_http_port(self):
        return 80

    def retrieve_gateway_network(self):
        return self.gateway_network

    def preflight(self, init: InitContext, dns: DnsService, name: NameService):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        self.public_ips = dns.get_dns_result().ip_addresses
        self.public_hosts = dns.get_dns_result().hostnames
        self.gateway_network = name.retrieve_name().replace("-", "_")

    def build(
        self,
        context: ExecutionContext,
    ):
        caddyfile = """
{
    auto_https off
}
        """

        for key, port in self.exposed_ports.items():
            if port.tls:
                caddyfile += f"""
:{port.port} {{
    tls /certs/caddy.crt /certs/caddy.key
    reverse_proxy {port.host}:{port.to}
}}
                """
            else:
                caddyfile += f"""
:{port.port} {{
    reverse_proxy {port.host}:{port.to}
}}
                """

        for protocol in ["http", "https"]:
            caddyfile += f"""
{protocol}:// {{
"""
            caddyfile += (
                """
    tls /certs/caddy.crt /certs/caddy.key
    """
                if protocol == "https"
                else ""
            )
            caddyfile += """
    header {
        -Server
        X-Forwarded-Proto {scheme}
        X-Forwarded-For {remote}
        X-Forwarded-Port {server_port}
        X-Forwarded-Host {host}
    }
        """

            for path_name, exposed_host in self.exposed_hosts.items():
                if exposed_host.stip_prefix:
                    caddyfile += f"""
    @{path_name} path /{path_name}* 
    handle @{path_name} {{
        uri strip_prefix /{path_name}
        reverse_proxy {exposed_host.host}:{exposed_host.port}
    }}
                """
                else:
                    caddyfile += f"""
    @{path_name} path /{path_name}*
    handle @{path_name} {{
        reverse_proxy {exposed_host.host}:{exposed_host.port}
    }}
                """

            for path_name, exposed_to_host in self.exposed_to_hosts.items():
                caddyfile += f"""
    @{path_name} path /{path_name}*
    handle @{path_name} {{
        rewrite * /{exposed_to_host.to}{{uri}}
        reverse_proxy {exposed_to_host.host}:{exposed_to_host.port}
    }}
                """

            if self.http_expose_default:
                caddyfile += f"""
    handle {{
        reverse_proxy {self.http_expose_default.host}:{self.http_expose_default.port}
    }}
            """

            caddyfile += """
}
        """

        context.file_tree.set_nested("configs", "Caddyfile", caddyfile)

        caddy_depends_on = []
        if self.with_certer:
            caddy_depends_on.append("certer")

        exposed_ports_strings = [
            f"{port.port}:{port.port}" for port in self.exposed_ports.values()
        ]

        caddy_container = {
            "image": "caddy:latest",
            "volumes": ["./configs/Caddyfile:/etc/caddy/Caddyfile", "./certs:/certs"],
            "ports": [
                f"{self.http_port}:80",
                f"{self.https_port}:443",
            ]
            + exposed_ports_strings,
            "depends_on": caddy_depends_on,
            "networks": [self.gateway_network, "default"],
        }

        context.docker_compose.set_nested("services", "caddy", caddy_container)

        context.file_tree.set_nested("certs", {})
        certer_container = {
            "image": self.certer_image,
            "volumes": ["./certs:/certs"],
            "environment": {
                "DOMAIN_NAMES": ",".join(self.public_hosts),
                "IP_ADDRESSED": ",".join(self.public_ips),
            },
        }

        context.docker_compose.set_nested("services", "certer", certer_container)
        context.docker_compose.set_nested(
            "networks",
            self.gateway_network,
            {"driver": "bridge", "name": self.gateway_network},
        )

    def expose(self, path_name: str, port: int, host: str, strip_prefix: bool = False):
        self.exposed_hosts[path_name] = ExposedHost(
            host=host, port=port, stip_prefix=strip_prefix
        )

    def expose_mapped(self, path_name: str, port: int, host: str, to: str):
        self.exposed_to_hosts[path_name] = ExpostedToHost(host=host, port=port, to=to)

    def expose_default(self, port: int, host: str):
        self.http_expose_default = ExposedHost(host=host, port=port, stip_prefix=False)

    def expose_port(self, port: int, host: str, tls: bool = False):
        self.exposed_ports[port] = ExposedPort(port=port, host=host, tls=tls, to=port)

    def expose_port_to(self, port: int, host: str, to_port: int, tls: bool = False):
        self.exposed_ports[port] = ExposedPort(
            port=port, host=host, tls=tls, to=to_port
        )

    def get_options(self):
        with_public_urls = Option(
            subcommand="public_url",
            help="Which public urls to use",
            type=str,
            multiple=True,
            default=DEFAULT_PUBLIC_URLS,
            show_default=True,
        )
        with_public_services = Option(
            subcommand="public_hosts",
            help="Which public hosts to use",
            type=str,
            multiple=True,
            default=DEFAULT_PUBLIC_HOSTS,
            show_default=True,
        )

        return [with_public_urls, with_public_services]
